/*
** Module   :COMPILE.CPP
** Abstract :Keyboard macro compiler
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Fri  27/03/1998     Created
*/

#include <string.h>

#include <boxcoll.h>
#include <version.h>

static int cx2n(char chr)
{
    int n = 0;

    if(__isdd(chr))
        return chr - '0';

    if(chr >= 'a' || chr <= 'z')
        return chr - 'a' + 0x0A;

    if(chr >= 'A' || chr <= 'Z')
        return chr -'A' + 0x0A;

    return 0;
}

int KeyDefCollection::compile_keydef(char *str, char *out)
{
    union
    {
        PEditProc proc;
        char data[sizeof(PEditProc)];
    }
    cvt;

    int cnt = 0;
    while(*str)
    {
        switch(*str)
        {
            case ' ':
            case ',':
            case '\r':
            case '\n':
            case '\t':
                break;

            case '"':
            case '{':
            case '\'':
                {
                    int iRexx = 0;

                    char delim = *str++;

                    if(delim == '{') //This is rexx?
                    {
                        delim = '}';

                        cnt += sizeof(cvt.data) + 1;

                        if(out)
                        {
                            cvt.proc = EditBoxCollection::doRexx;
                            *out++ = FUNC_ESCAPE;
                            memcpy(out, cvt.data, sizeof(cvt.data));
                            out += sizeof(cvt.data);
                        }

                        iRexx++;
                    }

                    while(*str)
                    {
                        if(delim != '}' && *str == delim)
                            break;

                        if(delim == '}' && *str == delim && !str[1])
                            break;

                        char chr = *str;

                        if(*str == '\\' && !iRexx) //Escape sequences
                        {
                            str++;

                            switch(*str)
                            {
                                case 'n':
                                    chr = '\n';
                                    break;

                                case 'r':
                                    chr = '\r';
                                    break;

                                case 't':
                                	chr = '\t';
                                    break;

                                case 'x':
                                    {
                                        str++;
                                        chr = 0;
                                        int k;

                                        for(k = 0; k < 2; k++)
                                        {
                                            if(!__ishd(*str))
                                                break;

                                            chr <<= 4;

                                            chr |= cx2n(*str++);
                                        }

                                        if(out)
                                            *out++ = chr++;
                                    }
                                    continue;

                                default:
                                	chr = *str;
                                	break;
                            }
                        }

                        if(out)
                            *out++ = chr;

                        cnt++;
                        str++;
                    }

                    if(*str && delim == '}')
                        return cnt;
                }
                break;

            default:
                if(__isis(*str))
                {
                    char *ptr = str;
                    while(*str && __isic(*str))
                        str++;

                    PDEntry pFunc = (PDEntry)Func_DIC.IsIn(ptr, str - ptr, 0);

                    if(pFunc)
                    {
                        cnt += sizeof(cvt.data) + 1; //function code

                        if(out)
                        {
                            cvt.proc = pFunc->rx.proc;
                            *out++ = FUNC_ESCAPE;
                            memcpy(out, cvt.data, sizeof(cvt.data));
                            out += sizeof(cvt.data);
                        }
                        if(!memcmp(pFunc->key, "markgo", 6) ||
                           !memcmp(pFunc->key, "markset", 7) ||
                           !memcmp(pFunc->key, "openjumplist", 12))
                        {
                            //Parameters
							cnt++;

							if(out)
							{
								char* num = pFunc->key + strlen(pFunc->key) - 1;

								*out++ = (*num - '0');
							}
                        }
                    }
                    else
                        return -2; //Wrong function name
                }
                else
                    return -1; //Wrong char in keydef
        }
        if(*str)
        	str++;
    }
    return cnt;
}

